from typing import Dict, List, Mapping, Set, Tuple, Union  # noqa

from uqbar.containers import UniqueTreeList

from .Attributes import Attributes
from .Edge import Edge  # noqa
from .Node import Node  # noqa


class Graph(UniqueTreeList):
    """
    A Graphviz graph, subgraph or cluster.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Core Classes"

    ### INITIALIZER ###

    def __init__(
        self,
        children=None,
        *,
        attributes: Union[Mapping[str, object], Attributes] = None,
        edge_attributes: Union[Mapping[str, object], Attributes] = None,
        is_cluster: bool = False,
        is_digraph: bool = True,
        name: str = None,
        node_attributes: Union[Mapping[str, object], Attributes] = None,
    ) -> None:
        UniqueTreeList.__init__(self, name=name, children=children)
        self._attributes = Attributes(
            "cluster" if is_cluster else "graph", **(attributes or {})
        )
        self._edge_attributes = Attributes("edge", **(edge_attributes or {}))
        self._node_attributes = Attributes("node", **(node_attributes or {}))
        self._is_cluster = bool(is_cluster)
        self._is_digraph = bool(is_digraph)

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: str = None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        def recurse(graph):
            indent = "    "
            result = []
            if not graph.parent:
                name = graph.name or "G"
                if graph.is_digraph:
                    string = "digraph {} {{".format(Attributes._format_value(name))
                else:
                    string = "graph {} {{".format(Attributes._format_value(name))
            else:
                if graph.name is not None:
                    name = graph.name
                    if graph.is_cluster:
                        name = "cluster_{}".format(name)
                else:
                    name = graph._get_canonical_name()
                string = "subgraph {} {{".format(Attributes._format_value(name))
            result.append(string)
            if graph.attributes:
                attributes = "graph {}".format(
                    format(graph.attributes, "graphviz")
                ).split("\n")
                result.extend(indent + line for line in attributes)
            if graph.node_attributes:
                attributes = "node {}".format(
                    format(graph.node_attributes, "graphviz")
                ).split("\n")
                result.extend(indent + line for line in attributes)
            if graph.edge_attributes:
                attributes = "edge {}".format(
                    format(graph.edge_attributes, "graphviz")
                ).split("\n")
                result.extend(indent + line for line in attributes)
            for child in graph:
                if isinstance(child, type(self)):
                    lines = (indent + line for line in recurse(child))
                else:
                    lines = (
                        indent + line for line in format(child, "graphviz").split("\n")
                    )
                result.extend(lines)
            if graph in edge_parents:
                for edge in edge_parents[graph]:
                    lines = format(edge, "graphviz").split("\n")
                    result.extend(indent + line for line in lines)
            result.append("}")
            return result

        all_edges: Set[Edge] = set()
        for child in self.depth_first():
            for edge in getattr(child, "edges", ()):
                if edge.tail.root is not edge.head.root:
                    continue
                all_edges.add(edge)
        edge_parents: Dict[Graph, List[Edge]] = {}
        for edge in sorted(
            all_edges, key=lambda edge: (edge.tail_graph_order, edge.head_graph_order)
        ):
            highest_parent = edge._get_highest_parent()
            edge_parents.setdefault(highest_parent, []).append(edge)

        return "\n".join(recurse(self))

    ### PRIVATE METHODS ###

    def _get_canonical_name(self) -> str:
        name_prefix = "graph"
        if self.is_cluster:
            name_prefix = "cluster"
        if self.name is not None:
            name = self.name
            root = self.root
            if root:
                instances = root[self.name]
                if not isinstance(instances, type(self)):
                    name = "{}_{}".format(name, instances.index(self))
            suffix = name
        elif self.graph_order:
            suffix = "_".join(str(x) for x in self.graph_order)
        else:
            suffix = "0"
        return "{}_{}".format(name_prefix, suffix)

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs

        return (uqbar.graphs.Graph, uqbar.graphs.Node)

    ### PUBLIC PROPERTIES ###

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @property
    def edge_attributes(self) -> Attributes:
        return self._edge_attributes

    @property
    def is_cluster(self) -> bool:
        return self._is_cluster

    @property
    def is_digraph(self) -> bool:
        return self._is_digraph

    @property
    def node_attributes(self) -> Attributes:
        return self._node_attributes



import copy

from uqbar.containers.UniqueTreeNode import UniqueTreeNode


class UniqueTreeContainer(UniqueTreeNode):

    ### INITIALIZER ###

    def __init__(self, children=None, name=None):
        UniqueTreeNode.__init__(self, name=name)
        self._named_children = {}

    ### SPECIAL METHODS ###

    def __contains__(self, expr):
        if isinstance(expr, str):
            return expr in self._named_children
        for x in self._children:
            if x is expr:
                return True
        return False

    def __iter__(self):
        for child in self._children:
            yield child

    def __len__(self):
        return len(self._children)

    ### PRIVATE METHODS ###

    def _cache_named_children(self):
        name_dictionary = super()._cache_named_children()
        if hasattr(self, "_named_children"):
            for name, children in self._named_children.items():
                name_dictionary[name] = copy.copy(children)
        return name_dictionary

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self):
        return UniqueTreeNode

    ### PUBLIC METHODS ###

    def depth_first(self, top_down=True, prototype=None):
        for child in tuple(self):
            if top_down:
                if not prototype or isinstance(child, prototype):
                    yield child
            if isinstance(child, UniqueTreeContainer):
                yield from child.depth_first(top_down=top_down, prototype=prototype)
            if not top_down:
                if not prototype or isinstance(child, prototype):
                    yield child

    def recurse(self, prototype=None):
        return self.depth_first(prototype=prototype)

    ### PUBLIC PROPERTIES ###

    @property
    def children(self):
        return tuple(self._children)



import abc
import argparse
import os
import pathlib
import sys
import typing
from configparser import ConfigParser

import uqbar.strings


class CLI(abc.ABC):
    """
    Abstract base for CLI scripts.
    """

    ### CLASS VARIABLES ###

    _colors = {
        "BLUE": "\033[94m",
        "RED": "\033[91m",
        "GREEN": "\033[92m",
        "END": "\033[0m",
    }

    config_name: str = ".uqbarrc"
    long_description: typing.Optional[str] = None
    short_description: typing.Optional[str] = None
    version: float = 1.0

    ### INITIALIZER ###

    def __init__(self):
        short_description = getattr(self, "short_description", None)
        long_description = getattr(self, "long_description", None)
        if long_description:
            long_description = uqbar.strings.normalize(long_description)
        parser = self._argument_parser = argparse.ArgumentParser(
            description=short_description,
            epilog=long_description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            prog=self.program_name,
        )
        version = "%(prog)s {}"
        version = version.format(getattr(self, "version", 1.0))
        parser.add_argument("--version", action="version", version=version)
        self._argument_parser = parser
        self._setup_argument_parser(parser)
        self._config_parser = None

    ### SPECIAL METHODS ###

    def __call__(self, arguments=None) -> None:
        """
        Calls developer script.

        Returns none.
        """
        self._config_parser = self._read_config_files()
        if arguments is None:
            arguments = self.argument_parser.parse_args()
        else:
            if isinstance(arguments, str):
                arguments = arguments.split()
            elif not isinstance(arguments, (list, tuple)):
                raise ValueError
            arguments = self.argument_parser.parse_args(arguments)
        self._process_args(arguments)
        sys.exit(0)

    ### PRIVATE METHODS ###

    def _is_valid_path(self, path) -> bool:
        path = pathlib.Path(path)
        return path.exists() and path.is_dir()

    @abc.abstractmethod
    def _process_args(self, arguments):
        raise NotImplementedError

    def _read_config_files(self) -> ConfigParser:
        paths = []
        home_config = pathlib.Path().home() / self.config_name
        if home_config.exists():
            paths.append(home_config)
        path = pathlib.Path.cwd()
        while not path.joinpath(self.config_name).exists():
            path = path.parent
            if path.parent == path:
                break
        if path.joinpath(self.config_name).exists():
            if path.joinpath(self.config_name) not in paths:
                paths.append(path.joinpath(self.config_name))
        config_parser = ConfigParser()
        config_parser.read([str(_) for _ in paths])
        return config_parser

    @abc.abstractmethod
    def _setup_argument_parser(self, parser: argparse.ArgumentParser):
        raise NotImplementedError

    def _validate_path(self, path):
        message = "{!r} is not a valid directory."
        message = message.format(path)
        error = argparse.ArgumentTypeError(message)
        path = os.path.abspath(path)
        if not self._is_valid_path(path):
            raise error
        return os.path.relpath(path)

    ### PUBLIC PROPERTIES ###

    @property
    def argument_parser(self):
        """
        The script's instance of argparse.ArgumentParser.
        """
        return self._argument_parser

    @property
    def formatted_help(self):
        """
        Formatted help of developer script.
        """
        return self._argument_parser.format_help()

    @property
    def formatted_usage(self):
        """
        Formatted usage of developer script.
        """
        return self._argument_parser.format_usage()

    @property
    def formatted_version(self):
        """
        Formatted version of developer script.
        """
        return self._argument_parser.format_version()

    @property
    def program_name(self):
        """
        The name of the script, callable from the command line.
        """
        name = "-".join(
            word.lower() for word in uqbar.strings.delimit_words(type(self).__name__)
        )
        return name



import enum
import inspect
from typing import List

from uqbar.apis.ClassDocumenter import ClassDocumenter


class SummarizingClassDocumenter(ClassDocumenter):
    """
    A summarizing class documenter.

    Organizes class members by category, separated by category title and
    horizontal rule.

    Categories include:

    -  Special methods
    -  Methods
    -  Class and static methods
    -  Read/write properties
    -  Read-only properties

    ::

        >>> import uqbar.apis
        >>> path = 'uqbar.apis.SummarizingClassDocumenter.SummarizingClassDocumenter'
        >>> documenter = uqbar.apis.SummarizingClassDocumenter(path)
        >>> documentation = str(documenter)
        >>> print(documentation)
        .. autoclass:: SummarizingClassDocumenter
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Attributes Summary
              :class: class-header
        <BLANKLINE>
           .. autosummary::
              :nosignatures:
        <BLANKLINE>
              __str__
              ignored_special_methods
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Special methods
              :class: class-header
        <BLANKLINE>
           .. automethod:: SummarizingClassDocumenter.__str__
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Class & static methods
              :class: class-header
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. automethod:: SummarizingClassDocumenter.validate_client
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Read-only properties
              :class: class-header
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. autoattribute:: SummarizingClassDocumenter.client
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. autoattribute:: SummarizingClassDocumenter.documentation_section
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. autoattribute:: SummarizingClassDocumenter.package_path

    :param package_path: the module path and name of the member to document
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Documenters"

    ignored_special_methods = (
        "__dict__",
        "__getattribute__",
        "__getnewargs__",
        "__getstate__",
        "__init__",
        "__reduce__",
        "__reduce_ex__",
        "__setstate__",
        "__sizeof__",
        "__subclasshook__",
        "fromkeys",
        "pipe_cloexec",
    )

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        name = getattr(self.client, "__name__")
        if issubclass(self.client, Exception):  # type: ignore
            return ".. autoexception:: {}".format(name)
        attributes = self._classify_class_attributes()
        (
            class_methods,
            data,
            methods,
            readonly_properties,
            readwrite_properties,
            special_methods,
            static_methods,
        ) = attributes
        result = [".. autoclass:: {}".format(name)]
        if issubclass(self.client, enum.Enum):  # type: ignore
            result.extend(["   :members:", "   :undoc-members:"])
        else:
            result.extend(self._build_member_autosummary(attributes))
        result.extend(
            self._build_attribute_section(
                special_methods, "automethod", "Special methods"
            )
        )
        result.extend(self._build_attribute_section(methods, "automethod", "Methods"))
        result.extend(
            self._build_attribute_section(
                sorted(class_methods + static_methods, key=lambda x: x.name),
                "automethod",
                "Class & static methods",
            )
        )
        result.extend(
            self._build_attribute_section(
                readwrite_properties, "autoattribute", "Read/write properties"
            )
        )
        result.extend(
            self._build_attribute_section(
                readonly_properties, "autoattribute", "Read-only properties"
            )
        )
        return "\n".join(result)

    ### PRIVATE METHODS ###

    def _build_attribute_section(
        self, attributes, directive: str, title: str
    ) -> List[str]:
        result: List[str] = []
        if not attributes:
            return result
        result.extend(
            [
                "",
                "   .. raw:: html",
                "",
                "      <hr/>",
                "",
                "   .. rubric:: {}".format(title),
                "      :class: class-header",
            ]
        )
        for attribute in attributes:
            result.append("")
            autodoc_directive = "   .. {}:: {}.{}".format(
                directive, getattr(self.client, "__name__"), attribute.name
            )
            if attribute.defining_class is self.client:
                result.append(autodoc_directive)
            else:
                result.append("   .. container:: inherited")
                result.append("")
                result.append("   {}".format(autodoc_directive))
        return result

    def _build_member_autosummary(self, attributes) -> List[str]:
        result: List[str] = []
        all_attributes: List[inspect.Attribute] = []
        for attribute_section in attributes:
            all_attributes.extend(
                attribute
                for attribute in attribute_section
                if attribute.defining_class is self.client
            )
        all_attributes.sort(key=lambda x: x.name)
        if not all_attributes:
            return result
        result.extend(
            [
                "",
                "   .. raw:: html",
                "",
                "      <hr/>",
                "",
                "   .. rubric:: {}".format("Attributes Summary"),
                "      :class: class-header",
                "",
                "   .. autosummary::",
                "      :nosignatures:",
                "",
            ]
        )
        for attribute in all_attributes:
            result.append("      {}".format(attribute.name))
        return result

    def _classify_class_attributes(self):
        class_methods = []
        data = []
        methods = []
        readonly_properties = []
        readwrite_properties = []
        special_methods = []
        static_methods = []
        attrs = inspect.classify_class_attrs(self.client)
        for attr in attrs:
            if attr.defining_class is object:
                continue
            elif (
                getattr(self.client, "__documentation_ignore_inherited__", None)
                and attr.defining_class is not self.client
            ):
                continue
            # Handle un-gettable attrs like Flask-SQLAlchemy's Model's `query`
            try:
                getattr(self.client, attr.name)
            except Exception:
                continue
            if attr.kind == "method":
                if attr.name not in self.ignored_special_methods:
                    if attr.name.startswith("__"):
                        special_methods.append(attr)
                    elif not attr.name.startswith("_"):
                        methods.append(attr)
            elif attr.kind == "class method":
                if attr.name not in self.ignored_special_methods:
                    if attr.name.startswith("__"):
                        special_methods.append(attr)
                    elif not attr.name.startswith("_"):
                        class_methods.append(attr)
            elif attr.kind == "static method":
                if attr.name not in self.ignored_special_methods:
                    if attr.name.startswith("__"):
                        special_methods.append(attr)
                    elif not attr.name.startswith("_"):
                        static_methods.append(attr)
            elif attr.kind == "property" and not attr.name.startswith("_"):
                if attr.object.fset is None:
                    readonly_properties.append(attr)
                else:
                    readwrite_properties.append(attr)
            elif (
                attr.kind == "data"
                and not attr.name.startswith("_")
                and attr.name not in getattr(self.client, "__slots__", ())
            ):
                data.append(attr)
        class_methods = tuple(sorted(class_methods))
        data = tuple(sorted(data))
        methods = tuple(sorted(methods))
        readonly_properties = tuple(sorted(readonly_properties))
        readwrite_properties = tuple(sorted(readwrite_properties))
        special_methods = tuple(sorted(special_methods))
        static_methods = tuple(sorted(static_methods))
        result = (
            class_methods,
            data,
            methods,
            readonly_properties,
            readwrite_properties,
            special_methods,
            static_methods,
        )
        return result



import threading

from .UniqueTreeContainer import UniqueTreeContainer


class UniqueTreeSet(UniqueTreeContainer):
    """
    A set-like node in a "unique" tree.
    """

    ### INITIALIZER ###

    def __init__(self, children=None, name=None):
        super().__init__(name=name)
        self._children = set()
        self._lock = threading.RLock()
        if children is not None:
            self.update(children)

    ### PRIVATE METHODS ###

    def _mutate(self, new_nodes, old_nodes):
        with self._lock:
            self._validate(new_nodes, old_nodes)
            self._update_parentage(new_nodes, old_nodes)
            self._mark_entire_tree_for_later_update()

    def _update_parentage(self, new_nodes, old_nodes):
        for old_node in old_nodes:
            old_node._set_parent(None)
        for new_node in new_nodes:
            new_node._set_parent(self)
        self._children.update(new_nodes)

    def _validate(self, new_nodes):
        parentage = self.parentage
        for new_node in new_nodes:
            if not isinstance(new_node, self._node_class):
                raise ValueError(f"Expected {self._node_class}, got {type(new_node)}")
            elif new_node in parentage:
                raise ValueError("Cannot set parent node as child.")

    ### PUBLIC METHODS ###

    def add(self, node):
        self._mutate(new_nodes=(node,), old_nodes=())

    def clear(self):
        self._mutate(new_nodes=(), old_nodes=self._children)

    def pop(self):
        with self._lock:
            node = self._children.pop()
            self._mutate(new_nodes=(), old_nodes=(node,))
        return node

    def remove(self, node):
        with self._lock:
            self._children.remove(node)
            self._mutate(new_nodes=(), old_nodes=(node,))

    def update(self, expr):
        with self._lock:
            self._mutate(new_nodes=expr, old_nodes=())
